# -*- coding: utf-8 -*-
#   Revenue Recognition for OpenERP
#   Copyright 2011 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Affero Public License version 3, as published 
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranties of 
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
#   PURPOSE.  See the GNU Affero Public License for more details.
#
#   You should have received a copy of the GNU Affero Public License along 
#   with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from osv import fields,osv
import decimal_precision as dp
from tools.translate import _
from dateutil.relativedelta import relativedelta
import datetime
import calendar
import netsvc


class account_revenue_recognition(osv.osv):
    _name = 'account.revenue.recognition'
    _description = 'Revenue/Prepayment Recognition'

    _columns = {
        'name': fields.char('Name', size=64, readonly=True, required=True),
        'invoice_line_id': fields.many2one('account.invoice.line', 'Invoice Line', readonly=True, required=True),
        'state': fields.selection([('draft','Unposted'), ('posted','Posted'), ('cancelled','Cancelled')], 'State', readonly=True),
        'date': fields.date('Date', readonly=True),
        'effective_date': fields.date('Effective Date', readonly=True),
        'company_id': fields.many2one('res.company', 'Company', readonly=True),
        'debit_account_id': fields.many2one('account.account', 'Debit Account', readonly=True),
        'credit_account_id': fields.many2one('account.account', 'Credit Account', readonly=True),
        'amount': fields.float('Amount', readonly=True, digits_compute=dp.get_precision('Account')),
        'move_id': fields.many2one('account.move', 'Journal Entry', readonly=True),
        'journal_id': fields.many2one('account.journal', 'Journal', readonly=True),
        }

    _defaults = {
        'state': 'draft',
        'date': datetime.date.today().isoformat(), 
        }

account_revenue_recognition()


class account_invoice_line(osv.osv):
    _inherit='account.invoice.line'
    _name='account.invoice.line'

    _columns = {
        'revenue_recognition': fields.boolean('Revenue/Prepayment Recognition?', help=_('Periodic contracts, such as support contracts, should have the revenue recognised as the work is done rather than immediately. The revenue will be moved into the Unearned Income or Prepaid account, and the daily pro-rata amount will be journaled at the last day of each period until the contract is complete.')),
        'contract_start_date': fields.date('Contract Start Date'),
        'contract_end_date': fields.date('Contract End Date'),
        'unearned_account_id': fields.many2one('account.account', 'Balance Sheet Account', help=_("The unearned income or prepaid account that the value of this invoice line will be journaled to (defaults from 'property_account_unearned' or 'property_account_prepaid')")),
        }


    def _default_account_get(self, cr, uid, context=None):
        """
        Get the default unearned income account from the property_account_unearned property for the company.
        """
        company_id = self.pool.get('res.company')._company_default_get(cr, uid, 'account.invoice.line', context=context)

        property_obj = self.pool.get('ir.property')
        if context.get('type', 'out_invoice')=='out_invoice':
            prop_unearned_id = property_obj.search(cr, uid, [('name','=','property_account_unearned'),('company_id','=',company_id)])
        else:
            prop_unearned_id = property_obj.search(cr, uid, [('name','=','property_account_prepaid'),('company_id','=',company_id)])

        if not prop_unearned_id:
            return False
        else:
            # Read the property to get the account id
            unearned_data = property_obj.read(cr, uid, prop_unearned_id, ['value_reference'], context=context)
            unearned_id = unearned_data and unearned_data[0].get('value_reference',None) and int(unearned_data[0]['value_reference'].split(',')[1]) or False
            return unearned_id

    _defaults = {
        'unearned_account_id': _default_account_get,
        }


    def onchange_start_date(self, cr, uid, ids, start_date, end_date):
        res = {}
        if start_date and not end_date:
            date = datetime.datetime.strptime(start_date, '%Y-%m-%d').date()
            res["value"] = {"contract_end_date": (date + relativedelta(years=1)).isoformat() }

        return res

account_invoice_line()


class account_invoice(osv.osv):
    _inherit='account.invoice'
    _name='account.invoice'

    def action_move_create(self, cr, uid, ids, *args):
        """
        This calls the finalize_invoice_move_lines and creates the moves and the
        revenue recognition schedule.
        """
        super(account_invoice,self).action_move_create(cr, uid, ids, args)
        
        # Process the invoices
        for inv in self.browse(cr, uid, ids):
            if inv.type in ('out_invoice','in_invoice'):
                # Process the customer invoice
                for line in inv.invoice_line:
                    self._process_invoice_line(cr, uid, line, inv)

        return True


    def _refund_cleanup_lines(self, cr, uid, lines):
        super_lines = super(account_invoice,self)._refund_cleanup_lines(cr, uid, lines)

        records = []
        for line in super_lines:
            # Change many2one property fields from tuple to just the ID
            rec = line[2]
            for field in ('unearned_account_id',):
                rec[field] = rec.get(field, False) and rec[field][0]
            records.append((0,0,rec))

        return records


    def action_cancel(self, cr, uid, ids, *args):
        """
        Extend the standard cancelation method to:
         * reverse the initial revenue recognition journal
         * create a reverse journal for the sum of the processed schedule records
         * cancel unprocessed revenue recognition schedule records
        """
        super(account_invoice,self).action_cancel(cr, uid, ids, args)

        # Cancel the invoice line revenue recognition records
        for inv in self.browse(cr, uid, ids):
            for line in inv.invoice_line:
                self._cancel_invoice_line(cr, uid, line, inv)

        return True


    def refund(self, cr, uid, ids, date=None, period_id=None, description=None, journal_id=None):
        # Create the refund(s)
        new_ids = super(account_invoice,self).refund(cr,uid,ids,date=date,period_id=period_id,description=description,journal_id=journal_id)

        # Create the reverse journals for the revenue recognition lines from the original invoice
        for inv in self.browse(cr, uid, ids):
            for line in inv.invoice_line:
                self._cancel_invoice_line(cr, uid, line, inv, date, period_id)

        return new_ids


    def _process_invoice_line(self, cr, uid, line, invoice):
        """
        Handle the creation of the revenue recognition schedule and the initial journal to move
        the invoice line amount into the Unearned Income account.
        """
        # Exit if revenue recognition is not required
        if not line.revenue_recognition:
            return

        # Validate the contract start and end date
        if not line.contract_end_date > line.contract_start_date:
            raise osv.except_osv(_("Error in Invoice Line '%s'" % line.name), _("The contract end date must be after the contract start date."))

        # Get the revenue recognition journal
        domain = [('name','=',_('Recognition Journal')),'|',('company_id','=',invoice.company_id.id),('company_id','=',False)]
        journal_ids = self.pool.get('account.journal').search(cr, uid, domain)
        if len(journal_ids)==0:
            raise osv.except_osv(_("Error in Invoice Line '%s'" % line.name), _("Cannot find the Recognition Journal for this company."))

        # Generate the revenue move into Unearned Income/Prepaid account
        self._create_revenue_move(cr, uid, line, invoice, journal_ids[0])

        # Generate the revenue recognition schedule
        self._create_schedule(cr, uid, line, invoice, journal_ids[0])


    def _amount_currency(self, cr, uid, line, invoice):
        """
        Calculate the invoice line net amount based on the company currency. This amount will be used
        for the journals.
        """
        curr_obj = self.pool.get('res.currency')
        amount = curr_obj.compute(cr, uid, invoice.currency_id.id, invoice.company_id.currency_id.id,
                                  line.price_subtotal, context={'date':invoice.date_invoice})
        return amount


    def _create_schedule(self, cr, uid, line, invoice, journal_id):
        """
        Generate the revenue recognition schedule records, based on the contract start and end
        and the net invoice line value.
        """
        # Calculate the contract length in days and daily rate (rounded)
        contract_start = datetime.datetime.strptime(line.contract_start_date, '%Y-%m-%d').date()
        contract_end = datetime.datetime.strptime(line.contract_end_date, '%Y-%m-%d').date()
        days = (contract_end - contract_start).days
        amount = self._amount_currency(cr, uid, line, invoice)
        day_rate = round(amount / days, 2)
        
        # Create revenue recognition as a list
        if contract_start > datetime.date.today():
            this_date = contract_start
        else:
            this_date = datetime.date.today()
        period_start_date = contract_start
        schedule = []
        while this_date<contract_end:
            # Calculate last day of month (or contract)
            day = calendar.monthrange(this_date.year, this_date.month)[1]
            last_day = datetime.date(this_date.year, this_date.month, day)
            if last_day > contract_end:
                last_day = contract_end

            # Calculate the pro-rata revenue for the last period
            delta_days = (last_day - period_start_date).days + 1
            period_amount = delta_days * day_rate

            # Save this scheduled amount
            schedule.append({'date': last_day,
                             'amount': period_amount})

            # Set up the next period
            period_start_date = last_day + datetime.timedelta(days=1)
            this_date = period_start_date

        # Change the last schedule amount to account for rounding
        sum = 0.0
        for x in range(0,len(schedule)-1):
            sum += schedule[x]['amount']
        if len(schedule)>0:
            schedule[-1]['amount'] = round(amount - sum, 2)
        else:
            # The contract is complete (start and end date in past), so post whole amount today
            schedule.append({'date': datetime.date.today(),
                             'amount': amount})
            
        # Define the debit and credit accounts for revenue/prepayment
        if invoice.type=='out_invoice':
            # Customer invoice
            debit_account_id = line.unearned_account_id.id
            credit_account_id = line.account_id.id
        else:
            # Supplier invoice
            debit_account_id = line.account_id.id
            credit_account_id = line.unearned_account_id.id

        revrec_obj = self.pool.get('account.revenue.recognition')
        for sch in schedule:
            # Create each schedule record
            record = {
                'name': invoice.move_id.name,
                'invoice_line_id': line.id,
                'effective_date': sch['date'],
                'company_id': line.company_id.id,
                'debit_account_id': debit_account_id,
                'credit_account_id': credit_account_id,
                'amount': sch['amount'],
                'journal_id': journal_id,
                }
            record = self.schedule_create_hook(record, invoice, line, sch)
            revrec_obj.create(cr, uid, record)


    def schedule_create_hook(self, record, invoice, line, sch):
        """
        A hook to allow the schedule that is created to be overriden by other modules.
        """
        return record


    def _create_revenue_move(self, cr, uid, line, invoice, journal_id):
        """
        Generate the general journal to move the revenue from the invoice line into the
        'Unearned Income' account.
        """
        # Set the debit and credit accounts based on the invoice type
        if invoice.type=='out_invoice':
            # Customer invoice
            debit_account_id = line.account_id.id
            credit_account_id = line.unearned_account_id.id
            move_name = 'Revenue Recognition'
        else:
            debit_account_id = line.unearned_account_id.id
            credit_account_id = line.account_id.id
            move_name = 'Prepayment'
        
        # Move the revenue into the Unearned Income account
        amount = self._amount_currency(cr, uid, line, invoice)
        move_lines = []
        ml_debit = self._move_line_create(invoice, journal_id)
        ml_debit.update({
                'name':'/',
                'account_id': debit_account_id,
                'debit': amount})
        move_lines.append((0,0,ml_debit))

        ml_credit = self._move_line_create(invoice, journal_id)
        ml_credit.update({
                'name':move_name,
                'account_id': credit_account_id,
                'credit': amount})
        move_lines.append((0,0,ml_credit))
        
        # Create the account move and post it
        move = {
            'ref': invoice.move_id.name,
            'line_id': move_lines,
            'journal_id': journal_id,
            'date': invoice.date_invoice,
            'period_id': invoice.period_id.id,
            }
        move = self.revenue_move_hook(move, invoice, line)

        context = {'journal_id': journal_id, 'period_id': invoice.period_id.id}
        move_id = self.pool.get('account.move').create(cr, uid, move, context=context)
        self.pool.get('account.move').post(cr, uid, [move_id], context={'invoice':invoice})

        # Create a revenue recognition record to record the move (so it can be tracked for reversing)
        rev_rec_obj = self.pool.get('account.revenue.recognition')
        record = {
            'name': invoice.move_id.name,
            'invoice_line_id': line.id,
            'effective_date': invoice.date_invoice,
            'company_id': line.company_id.id,
            'debit_account_id': debit_account_id,
            'credit_account_id': credit_account_id,
            'amount': amount,
            'move_id': move_id,
            'state': 'posted',
            'journal_id': journal_id,
            }
        record = self.revenue_schedule_hook(record, invoice, line)
        rev_rec_obj.create(cr, uid, record)


    def revenue_move_hook(self, record, invoice, line):
        """
        A hook that can be overriden by other modules to change the original
        revenue move before it is created.
        """
        return record


    def revenue_schedule_hook(self, record, invoice, line):
        """
        A hook that can be overriden by other modules to change the original
        revenue move before it is created.
        """
        return record

    
    def _move_line_create(self, invoice, journal_id):
        return {
            'date_maturity': invoice.date_invoice,
            'partner_id': invoice.partner_id.id,
            'date': invoice.date_invoice,
            'ref': invoice.move_id.name,
            'journal_id': journal_id,
            }

        
    def _period_get(self, cr, uid, company_id, date):
        # Get the current accounting period for the company
        period_ids = self.pool.get('account.period').search(cr, uid, [('date_start','<=',date),('date_stop','>=',date), ('company_id', '=', company_id)])
        if not period_ids or len(period_ids)==0:
            raise osv.except_osv(_('Error!'), _("Cannot find an Accounting Period for the company."))
        return period_ids[0]


    def post_revenue_recognition(self, cr, uid, *args):
        """
        Run by a scheduled task (daily).
        Reads the revenue recognition records for the run date (or older) and post financial journals for
        the draft schedule records.
        """
        logger = netsvc.Logger()
        logger.notifyChannel('post_revenue', netsvc.LOG_INFO,'Generating revenue recognition journals')
        
        rev_rec_obj = self.pool.get('account.revenue.recognition')

        # Get the scheduled revenue recognition records for today
        ids = rev_rec_obj.search(cr, uid, [('state','=','draft'),('effective_date','<=',datetime.date.today().isoformat())])

        for sch in rev_rec_obj.browse(cr, uid, ids):
            # Get the period for the company
            period_id = self._period_get(cr, uid, sch.company_id.id, sch.effective_date)

            # Create the general journal move lines
            lines = []
            credit = {
                'date_maturity': sch.effective_date,
                'partner_id': sch.invoice_line_id.invoice_id.partner_id.id,
                'date': sch.effective_date,
                'ref': sch.name,
                'journal_id': sch.journal_id.id,
                'name': 'Revenue Recognition/Prepayment',
                'account_id': sch.credit_account_id.id,
                'credit': sch.amount,
                }
            debit = {
                'date_maturity': sch.effective_date,
                'partner_id': sch.invoice_line_id.invoice_id.partner_id.id,
                'date': sch.effective_date,
                'ref': sch.name,
                'journal_id': sch.journal_id.id,
                'name': '/',
                'account_id': sch.debit_account_id.id,
                'debit': sch.amount,
                }
            lines.append((0,0,credit))
            lines.append((0,0,debit))

            # Create the move
            move = {
                'ref': sch.name,
                'line_id': lines,
                'journal_id': sch.journal_id.id,
                'date': sch.effective_date,
                'period_id': period_id,
                }
            move = self.post_move_hook(move, sch)

            context = {'journal_id': sch.journal_id.id, 'period_id': period_id}
            move_id = self.pool.get('account.move').create(cr, uid, move, context=context)

            # Post the general journal
            self.pool.get('account.move').post(cr, uid, [move_id], context=context)
          
            # Update the revenue recognition record to indicate that they have been posted
            rev_rec_obj.write(cr, uid, sch.id, {'state':'posted', 'move_id':move_id})

        logger.notifyChannel('post_revenue', netsvc.LOG_INFO,'Completed revenue recognition journals')


    def post_move_hook(self, move, schedule):
        """
        A hook that can be overriden to allow other modules to change the move that
        will be created for each schedule.
        """
        return move
    

    def _cancel_invoice_line(self, cr, uid, line, invoice, ondate=None, period_id=None):
        """
        Clean-up the revenue recognition schedule records when an invoice has been cancelled.
        Extend the standard cancelation method to:
         * reverse the initial revenue recognition journal
         * create a reverse journal for the sum of the processed schedule records
         * cancel unprocessed revenue recognition schedule records
        """
        # Exit if revenue recognition is not required
        if not line.revenue_recognition:
            return

        # Reverse processed schedule records
        rev_rec_obj = self.pool.get('account.revenue.recognition')
        posted_ids = rev_rec_obj.search(cr, uid, [('state','=','posted'),('invoice_line_id','=',line.id)])
        for rec in rev_rec_obj.browse(cr, uid, posted_ids):
            self._reverse_schedule_journal(cr, uid, rec, ondate, period_id)

        # Cancel all schedule records for this invoice line
        sch_ids = rev_rec_obj.search(cr, uid, [('invoice_line_id','=',line.id)])
        rev_rec_obj.write(cr, uid, sch_ids, {'state':'cancelled'})


    def _reverse_schedule_journal(self, cr, uid, rev_rec, ondate=None, period_id=None):
        """
        Reverse the journal linked to a posted revenue recognition record.
        """
        move_obj = self.pool.get('account.move')

        # Set the date of the reverse journal
        if not ondate:
            ondate = datetime.date.today().isoformat()

        # Get the period for the move
        if not period_id:
            period_id = self._get_period_for_date(cr, uid, ondate, rev_rec.move_id.company_id.id)
            if not period_id:
                raise osv.except_osv(_('UserError'),
                        _("Cannot find the Accounting Period for the date '%s'" % ondate))

        # Copy the move record
        copy_id = move_obj.copy(cr, uid, rev_rec.move_id.id)

        # Build the move update
        update_rec = {
            'name': '%s Cancel' % (rev_rec.move_id.name),
            'date': ondate,
            'period_id': period_id,
            }

        # Reverse the debit/credit on the move lines
        lines = []
        for line in move_obj.browse(cr, uid, copy_id).line_id:
            lines.append((1, line.id,
                          {
                           'credit': line.debit, 
                           'debit': line.credit, 
                           'date': ondate,
                           'period_id': period_id
                           })
                         )
        update_rec['line_id'] = lines

        # Update the move and lines
        move_obj.write(cr, uid, [copy_id], update_rec)

        for line in move_obj.browse(cr, uid, copy_id).line_id:
            line.write({
                    'credit': line.debit, 
                    'debit': line.credit, 
                    'date_maturity': ondate,
                    'date': ondate})

        # Post the reversed journal
        move_obj.post(cr, uid, [copy_id])


    def _get_period_for_date(self, cr, uid, date, company_id):
        period_ids = self.pool.get('account.period').search(cr, uid, [('date_start','<=',date),('date_stop','>=',date), ('company_id', '=', company_id)])
        if period_ids:
            return period_ids[0]
        else:
            return False

account_invoice()

